from enum import Enum
from ipaddress import IPv4Address, IPv4Network

from pydantic import Field
from sms_broker.server.config import BaseServerConfig


class UiServerConfig(BaseServerConfig):
    class Type(str, Enum):
        UI = "ui"

    type: Type = Field(default=..., title="Type", description="Type of the server.")
    host: IPv4Address = Field(default=..., title="UI Server Host", description="UI Server Host to bind to.")
    port: int = Field(default=..., title="UI Server Port", ge=0, le=65535, description="UI Server Port to bind to.")
    allowed_networks: list[IPv4Network] = Field(default_factory=lambda: [IPv4Network("0.0.0.0/0")], title="UI Server Allowed Clients Networks",
                                                description="List of allowed client networks.")
    authentication_enabled: bool = Field(default=False, title="UI Server Authentication Enabled", description="Enable UI Server Authentication.")
    authentication_accounts: dict[str, str] = Field(default_factory=dict, title="UI Server Authentication Accounts", description="UI Server Authentication Accounts.")
    session_secret_key: str = Field(default=..., title="UI Server Session Secret Key", description="UI Server Session Secret Key.", min_length=64)
    session_cookie: str = Field(default="session", title="UI Server Session Cookie", description="UI Server Session Cookie.")
    session_max_age: int = Field(default=60 * 60 * 24 * 14, title="UI Server Session Max Age", description="UI Server Session Max Age.")



import json
import logging
from contextlib import asynccontextmanager
from datetime import datetime
from enum import Enum
from ipaddress import IPv4Address
from json import JSONDecodeError
from typing import TYPE_CHECKING, Annotated, Any

import uvicorn
from pydantic import BaseModel, Field

from fastapi import Depends, HTTPException
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm

from sms_broker.statics import ASSETS_PATH
from sms_broker.db import Sms, SmsStatus
from sms_broker.server.server import BaseServer
from starlette.requests import Request

from sms_broker.settings import settings
from wiederverwendbar.fastapi import FastAPI

if TYPE_CHECKING:
    from sms_broker.server.api.config import ApiServerConfig

logger = logging.getLogger(__name__)


class SmsSendApiModel(BaseModel):
    error: bool = Field(default=False, title="Error", description="Indicates if the request resulted in an error.")
    sms_id: int | None = Field(default=None, title="SMS ID", description="The ID of the queued SMS.")
    result: str = Field(default=..., title="Message", description="The message of the response.")


class SmsStatusApiModel(BaseModel):
    id: int = Field(default=..., title="ID", description="The ID of the SMS.")
    status: SmsStatus = Field(default=..., title="Status", description="The status of the SMS.")
    received_by: str = Field(default=..., title="Received by", description="The IP address of the server that received the SMS.")
    received_datetime: datetime = Field(default=..., title="Received datetime", description="The datetime when the SMS was received.")
    processed_datetime: datetime | None = Field(default=None, title="Processed datetime", description="The datetime when the SMS was processed.")
    sent_by: str | None = Field(default=None, title="Sent by", description="The IP address of the server that sent the SMS.")
    number: str = Field(default=..., title="Number", description="The phone number of the SMS.")
    message: str = Field(default=..., title="Message", description="The message of the SMS.")
    result: str | None = Field(default=None, title="Result", description="The result of the SMS.")
    log: str | None = Field(default=None, title="Log", description="The log of the SMS.")


class ListOrderBy(str, Enum):
    ID = "id"
    STATUS = "status"
    RECEIVED_BY = "received_by"
    RECEIVED_DATETIME = "received_datetime"
    PROCESSED_DATETIME = "processed_datetime"
    SENT_BY = "sent_by"
    NUMBER = "number"
    MESSAGE = "message"
    RESULT = "result"


class ListOrderDesc(str, Enum):
    ASC = "asc"
    DESC = "desc"


class Unauthorized(BaseModel):
    detail: str

    class Config:
        json_schema_extra = {
            "example": {"detail": "Not authenticated"}
        }


class SmsNotFoundError(BaseModel):
    detail: str

    class Config:
        json_schema_extra = {
            "example": {"detail": "SMS with id=1 not found!"},
        }


class SmsAbortError(BaseModel):
    detail: str

    class Config:
        json_schema_extra = {
            "example": {"detail": "Cannot abort SMS with id=1!"},
        }


class ApiServer(BaseServer, FastAPI):
    __str_columns__ = ["name",
                       ("debug", "config_debug"),
                       ("host", "config_host"),
                       ("port", "config_port"),
                       ("allowed_networks", "config_allowed_networks"),
                       ("authentication_enabled", "config_authentication_enabled")]

    def __init__(self, name: str, config: "ApiServerConfig"):
        BaseServer.__init__(self, name=name, config=config)
        FastAPI.__init__(self,
                         lifespan=self._stated_done,
                         debug=self.config.debug,
                         title=f"{settings.branding_title} - {self.name}",
                         description=settings.branding_description,
                         version=f"v{settings.branding_version}",
                         terms_of_service=settings.branding_terms_of_service,
                         favicon=ASSETS_PATH / "favicon.ico",
                         docs_url=self.config.docs_web_path,
                         redoc_url=self.config.redoc_web_path,
                         contact={"name": settings.branding_author, "email": settings.branding_author_email},
                         license_info={"name": settings.branding_license, "url": settings.branding_license_url})

        self.init_done()

    def setup(self) -> None:
        super().setup()

        # async def validate_token() -> tuple[str, str]:
        #     return await self.get_api_credentials_from_token(token=token)

        if self.config.authentication_enabled:
            @self.post(path="/auth",
                       summary="Authenticate against server.",
                       tags=["API version 1"])
            async def auth(form_data: Annotated[OAuth2PasswordRequestForm, Depends()]):
                return await self.auth(form_data=form_data)

            @self.get(path="/sms",
                      summary="List SMS.",
                      tags=["API version 1"],
                      responses={401: {"model": Unauthorized}})
            async def list_sms(order_by: ListOrderBy = ListOrderBy.ID,
                               order_desc: ListOrderDesc = ListOrderDesc.ASC,
                               limit: int = 100,
                               page: int = 1,
                               query: str = "{}",
                               _=Depends(self.get_api_credentials_from_token)) -> list[
                SmsStatusApiModel]:
                return await self.list_sms(order_by=order_by,
                                           order_desc=order_desc,
                                           limit=limit,
                                           page=page,
                                           query=query)

            @self.get(path="/sms/{sms_id}",
                      summary="Get SMS.",
                      tags=["API version 1"],
                      responses={401: {"model": Unauthorized},
                                 404: {"model": SmsNotFoundError}})
            async def get_sms(sms_id: int,
                              _=Depends(self.get_api_credentials_from_token)) -> SmsStatusApiModel:
                return await self.get_sms(sms_id=sms_id)

            @self.post(path="/sms",
                       summary="Sending an SMS.",
                       tags=["API version 1"],
                       responses={401: {"model": Unauthorized}})
            async def send_sms(request: Request,
                               number: str,
                               message: str,
                               _=Depends(self.get_api_credentials_from_token)) -> SmsSendApiModel:
                return await self.send_sms(request=request,
                                           number=number,
                                           message=message)

            @self.patch(path="/sms/{sms_id}",
                        summary="Reset an SMS.",
                        tags=["API version 1"],
                        responses={401: {"model": Unauthorized},
                                   404: {"model": SmsNotFoundError}})
            async def reset_sms(sms_id: int,
                                _=Depends(self.get_api_credentials_from_token)) -> SmsStatusApiModel:
                return await self.reset_sms(sms_id=sms_id)

            @self.delete(path="/sms/{sms_id}",
                         summary="Abort an SMS.",
                         tags=["API version 1"],
                         responses={401: {"model": Unauthorized},
                                    403: {"model": SmsAbortError,
                                          "description": "Aborting SMS is not allowed."},
                                    404: {"model": SmsNotFoundError}})
            async def abort_sms(sms_id: int,
                                _=Depends(self.get_api_credentials_from_token)) -> SmsStatusApiModel:
                return await self.abort_sms(sms_id=sms_id)
        else:
            @self.get(path="/sms",
                      summary="List SMS.",
                      tags=["API version 1"],
                      responses={401: {"model": Unauthorized}})
            async def list_sms(order_by: ListOrderBy = ListOrderBy.ID,
                               order_desc: ListOrderDesc = ListOrderDesc.ASC,
                               limit: int = 100,
                               page: int = 1,
                               query: str = "{}") -> list[SmsStatusApiModel]:
                return await self.list_sms(order_by=order_by,
                                           order_desc=order_desc,
                                           limit=limit,
                                           page=page,
                                           query=query)

            @self.get(path="/sms/{sms_id}",
                      summary="Get SMS.",
                      tags=["API version 1"],
                      responses={401: {"model": Unauthorized},
                                 404: {"model": SmsNotFoundError}})
            async def get_sms(sms_id: int) -> SmsStatusApiModel:
                return await self.get_sms(sms_id=sms_id)

            @self.post(path="/sms",
                       summary="Sending an SMS.",
                       tags=["API version 1"],
                       responses={401: {"model": Unauthorized}})
            async def send_sms(request: Request,
                               number: str,
                               message: str) -> SmsSendApiModel:
                return await self.send_sms(request=request,
                                           number=number,
                                           message=message)

            @self.patch(path="/sms/{sms_id}",
                        summary="Reset an SMS.",
                        tags=["API version 1"],
                        responses={401: {"model": Unauthorized},
                                   404: {"model": SmsNotFoundError}})
            async def reset_sms(sms_id: int) -> SmsStatusApiModel:
                return await self.reset_sms(sms_id=sms_id)

            @self.delete(path="/sms/{sms_id}",
                         summary="Abort an SMS.",
                         tags=["API version 1"],
                         responses={401: {"model": Unauthorized},
                                    403: {"model": SmsAbortError,
                                          "description": "Aborting SMS is not allowed."},
                                    404: {"model": SmsNotFoundError}})
            async def abort_sms(sms_id: int) -> SmsStatusApiModel:
                return await self.abort_sms(sms_id=sms_id)

    @property
    def config(self) -> "ApiServerConfig":
        return super().config

    @property
    def config_host(self) -> str:
        return str(self.config.host)

    @property
    def config_port(self) -> int:
        return self.config.port

    @property
    def config_allowed_networks(self) -> list[str]:
        return [str(allowed_network) for allowed_network in self.config.allowed_networks]

    @property
    def config_authentication_enabled(self) -> bool:
        return self.config.authentication_enabled

    @staticmethod
    @asynccontextmanager
    async def _stated_done(api_server: "ApiServer"):
        api_server.stated_done()
        yield

    def enter(self):
        uvicorn.run(self, host=str(self.config.host), port=self.config.port)

    def exit(self):
        ...

    async def get_api_credentials_from_token(self, token: Annotated[str, Depends(OAuth2PasswordBearer(tokenUrl="/auth"))]) -> tuple[str, str]:
        if not self.config.authentication_enabled:
            raise HTTPException(status_code=401, detail="Authentication is disabled.")
        if ":" not in token:
            raise HTTPException(status_code=401, detail="Invalid authentication credentials")
        split_token = token.split(":")
        if len(split_token) != 2:
            raise HTTPException(status_code=401, detail="Invalid authentication credentials")
        api_key, api_secret = split_token
        if api_key not in self.config.authentication_accounts:
            raise HTTPException(status_code=401, detail="API key not found.")
        if api_secret != self.config.authentication_accounts[api_key]:
            raise HTTPException(status_code=401, detail="API secret is incorrect.")
        return api_key, api_secret

    # noinspection DuplicatedCode
    def handle_request(self, caller: None, **kwargs) -> Any | None:
        # check if client ip is allowed
        allowed = False
        for network in self.config.allowed_networks:
            if kwargs["client_ip"] in network:
                allowed = True
                break
        if not allowed:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=False, sms_id=None, result=f"Client IP address '{kwargs['client_ip']}' is not allowed.")

        logger.debug(f"{self} - Accept message:\nclient='{kwargs['client_ip']}'\nport={kwargs['client_ip']}")

        return super().handle_request(caller=caller, **kwargs)

    def handle_sms_data(self, caller: None, **kwargs) -> tuple[str, str]:
        return kwargs["number"], kwargs["message"]

    def success_handler(self, caller: None, sms_id: int, result: str, **kwargs) -> Any:
        if self.config.success_result is not None:
            result = self.config.success_result
        return SmsSendApiModel(error=False, sms_id=sms_id, result=result)

    def error_handler(self, caller: None, sms_id: int | None, result: str, **kwargs) -> Any:
        if self.config.error_result is not None:
            result = self.config.error_result
        return SmsSendApiModel(error=True, sms_id=sms_id, result=result)

    async def auth(self, form_data: Annotated[OAuth2PasswordRequestForm, Depends()]):
        if not self.config.authentication_enabled:
            raise HTTPException(status_code=401, detail="Authentication is disabled.")
        if form_data.username not in self.config.authentication_accounts:
            raise HTTPException(status_code=401, detail="API key not found.")
        if form_data.password != self.config.authentication_accounts[form_data.username]:
            raise HTTPException(status_code=401, detail="API secret is incorrect.")
        return {"access_token": f"{form_data.username}:{form_data.password}", "token_type": "bearer"}

    async def list_sms(self,
                       order_by: ListOrderBy = ListOrderBy.ID,
                       order_desc: ListOrderDesc = ListOrderDesc.ASC,
                       limit: int = 100,
                       page: int = 1,
                       query: str = "{}") -> list[SmsStatusApiModel]:
        # try to parse query
        try:
            query_dict = json.loads(query)
        except JSONDecodeError as e:
            msg = f"Error while parsing query!"
            if self.config.debug:
                msg = f"{msg[:-1]}:\n{e}"
            logger.error(msg)
            raise HTTPException(status_code=442, detail=[{"loc": "query", "msg": msg}])

        # list from db
        sms_models: list[SmsStatusApiModel] = []
        for sms_dict in Sms.get_all(order_by=order_by.value, order_desc=order_desc is ListOrderDesc.DESC, rows_per_page=limit, page=page, as_dict=True, **query_dict):
            sms_model = SmsStatusApiModel(**sms_dict)
            sms_models.append(sms_model)
        return sms_models

    async def get_sms(self,
                      sms_id: int) -> SmsStatusApiModel:
        sms_dict = Sms.get(Sms.id == sms_id, as_dict=True)
        if sms_dict is None:
            raise HTTPException(status_code=404, detail=f"SMS with id={sms_id} not found.")
        sms_model = SmsStatusApiModel(**sms_dict)
        return sms_model

    async def send_sms(self,
                       request: Request,
                       number: str,
                       message: str) -> SmsSendApiModel:
        # get client_ip and client_port
        try:
            client_ip = IPv4Address(request.client.host)
            client_port = request.client.port
        except Exception as e:
            self.handle_response(caller=self, log_level=logging.ERROR, success=e, sms_id=None, result=f"Error while parsing client IP address.")
            raise RuntimeError("This should never happen.")
        return self.handle_request(caller=None, number=number, message=message, client_ip=client_ip, client_port=client_port)

    async def reset_sms(self,
                        sms_id: int) -> SmsStatusApiModel:
        sms = Sms.get(Sms.id == sms_id)
        if sms is None:
            raise HTTPException(status_code=404, detail=f"SMS with id={sms_id} not found.")
        sms.update(status=SmsStatus.QUEUED,
                   processed_datetime=None,
                   sent_by=None,
                   result=None,
                   log=None)
        return await self.get_sms(sms_id=sms_id)

    async def abort_sms(self,
                        sms_id: int) -> SmsStatusApiModel:
        sms = Sms.get(Sms.id == sms_id)
        if sms is None:
            raise HTTPException(status_code=404, detail=f"SMS with id={sms_id} not found.")
        if sms.status != SmsStatus.QUEUED:
            raise HTTPException(status_code=403, detail=f"Cannot abort SMS with id={sms_id}! SMS is not in queued state.")
        sms.update(status=SmsStatus.ABORTED,
                   processed_datetime=None,
                   sent_by=None,
                   result=None,
                   log=None)
        return await self.get_sms(sms_id=sms_id)



import logging
import logging.handlers
import sys
from datetime import datetime, timedelta
from typing import Callable

from sqlalchemy import and_, or_
from wiederverwendbar.logger import LoggerSingleton, LoggingContext, remove_logger
from wiederverwendbar.task_manger import TaskManager, Task, EverySeconds

from sms_broker.db import Sms, SmsStatus
from sms_broker.settings import settings
from sms_broker.gateways.gateway import BaseGateway
from sms_broker.gateways.teltonika.gateway import TeltonikaGateway
from sms_broker.gateways.vonage.gateway import VonageGateway

IGNORED_LOGGERS_LIKE = ["sqlalchemy", "pymysql"]
# noinspection PyArgumentList
logger = LoggerSingleton(name=__name__,
                         settings=settings.worker,
                         ignored_loggers_like=IGNORED_LOGGERS_LIKE,
                         init=True)


class SmsWorker(TaskManager):
    class SmsLogHandler(logging.handlers.BufferingHandler):
        def __init__(self, buffer_target: Callable):
            super().__init__(capacity=10)
            self.buffer_target = buffer_target

        def flush(self):
            for record in self.buffer:
                self.buffer_target(self.format(record))
            super().flush()

    def __init__(self):
        logger.info(f"Initializing SMS-Worker ...")
        super().__init__(name="SMS-Worker", worker_count=settings.worker.count, daemon=True, logger=logger)

        # initialize gateway
        logger.info("Initializing gateways ...")
        self._next_sms_gateway_index: int | None = None
        self._gateways: list[BaseGateway] = []
        for gateway_config_name, gateway_config in settings.worker.gateways.items():
            if len(gateway_config_name) > 20:
                logger.error(f"Gateway name '{gateway_config_name}' is too long. Max size is 20 characters.")
                sys.exit(1)

            gateway_cls = None
            if gateway_config.type == "teltonika":
                gateway_cls = TeltonikaGateway
            elif gateway_config.type == "vonage":
                gateway_cls = VonageGateway

            if gateway_cls is None:
                logger.error(f"Unknown gateway type '{gateway_config.type}'.")
                sys.exit(1)

            gateway = gateway_cls(name=gateway_config_name, config=gateway_config)
            self._gateways.append(gateway)
        if len(self._gateways) == 0:
            logger.error(f"No gateways are configured. Please check your settings.")
            sys.exit(1)
        logger.debug("Initializing gateways ... done")

        # create tasks
        logger.info("Initializing tasks ...")
        Task(name="Handle SMS", manager=self, trigger=EverySeconds(settings.worker.sms_handle_interval), payload=self.handle_sms)
        Task(name="Cleanup SMS", manager=self, trigger=EverySeconds(settings.worker.sms_cleanup_interval), payload=self.cleanup_sms)
        logger.debug("Initializing tasks ... done")

        logger.debug(f"Initializing SMS-Worker ... done")

        logger.info(f"Starting SMS-Worker ...")

        # starting task manager workers
        self.start()

        logger.debug(f"Starting SMS-Worker ... done")

    def handle_sms(self):
        sms: Sms | None = None

        def get_sms() -> Sms | None:
            nonlocal sms
            sms = Sms.get(status=SmsStatus.QUEUED)
            return sms

        while get_sms():
            try:
                logger.info(f"Processing SMS with id={sms.id} ...")

                # calculate next_sms_gateway_index
                if self._next_sms_gateway_index is None:
                    self._next_sms_gateway_index = 0
                else:
                    self._next_sms_gateway_index += 1
                if self._next_sms_gateway_index >= len(self._gateways):
                    self._next_sms_gateway_index = 0

                # order gateways
                gateways: list[BaseGateway] = []
                for gateway in self._gateways[self._next_sms_gateway_index:]:
                    gateways.append(gateway)
                if self._next_sms_gateway_index > 0:
                    for gateway in self._gateways[:self._next_sms_gateway_index]:
                        gateways.append(gateway)

                # create sms logger
                sms_log = ""

                def add_sms_log(sms_log_msg: str):
                    nonlocal sms_log
                    if len(sms_log) > 0:
                        sms_log += "\n"
                    sms_log += sms_log_msg

                sms_logger = logging.Logger(name=f"{logger.name}", )
                sms_logger_handler = self.SmsLogHandler(buffer_target=add_sms_log)
                sms_logger_handler.setLevel(settings.worker.log_level.get_level_number())
                sms_logger_handler.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
                sms_logger.addHandler(sms_logger_handler)
                sms_logger.setLevel(settings.worker.log_level.get_level_number())

                # send sms with gateways
                with LoggingContext(sms_logger, ignore_loggers_like=IGNORED_LOGGERS_LIKE):
                    log_level = logging.ERROR
                    result = "Error while sending SMS. Not gateways left."
                    status = SmsStatus.ERROR
                    send_by = None
                    for gateway in gateways:
                        gateway.increase_sms_count()
                        # check if gateway is available
                        if not gateway.check():
                            gateway.increase_sms_error_count()
                            continue

                        # send it with gateway
                        success, log_level, result = gateway.send_sms(sms.number, sms.message)
                        if success:
                            status = SmsStatus.SENT
                            send_by = gateway.name
                            break
                        gateway.increase_sms_error_count()
                    sms_logger.log(log_level, result)
                    logger.log(log_level, result)

                # flush sms_logger_handler
                sms_logger_handler.flush()

                # remove logger
                remove_logger(sms_logger)

                # update sms
                sms.update(status=status,
                           processed_datetime=datetime.now(),
                           sent_by=send_by,
                           result=result,
                           log=sms_log)

                logger.debug(f"Processing SMS with id={sms.id} ... done")
            except Exception as e:
                logger.error(f"Error while processing SMS with id={sms.id}.\nException: {e}")

    def cleanup_sms(self):
        try:
            cleanup_datetime = datetime.now() - timedelta(seconds=settings.worker.sms_cleanup_max_age)
            expression = and_(Sms.status != SmsStatus.QUEUED,
                              or_(
                                  and_(
                                      Sms.received_datetime <= cleanup_datetime,
                                      Sms.processed_datetime is None
                                  ),
                                  Sms.processed_datetime <= cleanup_datetime
                              ))
            cleanup_sms_count = Sms.length(expression)
            if cleanup_sms_count == 0:
                return
            logger.info(f"Cleaning up {cleanup_sms_count} SMS ...")
            Sms.delete_all(expression)
            logger.debug(f"Cleaning up {cleanup_sms_count} SMS ... done")
        except Exception as e:
            logger.error(f"Error while getting SMS to cleanup.\nException: {e}")



import sys
import time

from wiederverwendbar.logger import LoggerSingleton

from sms_broker.settings import settings
from sms_broker.server.server import BaseServer
from sms_broker.server.file.server import FileServer
from sms_broker.server.tcp.server import TcpServer
from sms_broker.server.api.server import ApiServer
from sms_broker.server.ui.server import UiServer

IGNORED_LOGGERS_LIKE = ["sqlalchemy", "pymysql", "asyncio", "parso", "engineio", "socketio", "python_multipart.multipart"]
# noinspection PyArgumentList
logger = LoggerSingleton(name=__name__,
                         settings=settings.listener,
                         ignored_loggers_like=IGNORED_LOGGERS_LIKE,
                         init=True)


class SmsListener:
    def __init__(self):
        logger.info(f"Initializing SMS-Listener ...")

        # initialize servers
        logger.info("Initializing servers ...")
        self._server: list[BaseServer] = []
        for server_config_name, server_config in settings.listener.server.items():
            if len(server_config_name) > 20:
                logger.error(f"Server name '{server_config_name}' is too long. Max size is 20 characters.")
                sys.exit(1)

            server_cls = None
            if server_config.type == "file":
                server_cls = FileServer
            elif server_config.type == "tcp":
                server_cls = TcpServer
            elif server_config.type == "api":
                server_cls = ApiServer
            elif server_config.type == "ui":
                server_cls = UiServer

            if server_cls is None:
                logger.error(f"Unknown server type '{server_config.type}'.")
                sys.exit(1)

            server = server_cls(name=server_config_name, config=server_config)
            self._server.append(server)
        if len(self._server) == 0:
            logger.error(f"No servers are configured. Please check your settings.")
            sys.exit(1)
        logger.debug("Initializing servers ... done")

        logger.debug(f"Initializing SMS-Listener ... done")

        logger.info(f"Starting SMS-Listener ...")
        # starting servers
        for server in self._server:
            server.start()

        # waiting for servers to start
        while True:
            if not any(not server.is_started for server in self._server):
                break
            logger.debug(f"Waiting for servers to start ...")
            time.sleep(1.0)

        logger.debug(f"Starting SMS-Listener ... done")



import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Union, Any

from sms_broker.base import Base

if TYPE_CHECKING:
    from sms_broker.gateways.config import BaseGatewayConfig

logger = logging.getLogger(__name__)


class BaseGateway(Base):
    __str_columns__ = ["name",
                       ("dry_run", "config_dry_run")]

    def __init__(self, name: str, config: "BaseGatewayConfig"):
        Base.__init__(self, name=name, config=config)

        self._state = False

        self.init_done()

    @property
    def config(self) -> Union["BaseGatewayConfig", Any]:
        return super().config

    @property
    def config_dry_run(self) -> bool:
        return self.config.dry_run

    @property
    def state(self) -> bool:
        return self._state

    @state.setter
    def state(self, value: bool):
        self._state = value

    def check(self) -> bool:
        if not self._config.check:
            logger.warning(f"Gateway check is disabled for {self}. This is not recommended!")
            self.state = True
            return True
        try:
            logger.debug(f"Checking gateway {self} ...")
            if self._check():
                logger.debug(f"Gateway {self} is available.")
                self.state = True
                return True
            logger.warning(f"Gateway {self} is not available.")
        except Exception as e:
            logger.error(f"Gateway {self} check failed.\nException: {e}")
        self.state = False
        return False

    @abstractmethod
    def _check(self) -> bool:
        ...

    def send_sms(self, number: str, message: str) -> tuple[bool, int, str]:
        logger.debug(f"Sending SMS via {self} ...")
        log_level = logging.DEBUG
        success = False
        try:
            if not self.state:
                raise RuntimeError(f"SMS gateway {self} is not available!")
            if self._config.dry_run:
                log_level = logging.WARNING
                success = True
                result = f"Dry run mode is enabled. SMS will not sent via {self}."
            else:
                success, gateway_result = self._send_sms(number, message)
                if success:
                    result = f"SMS sent successfully via {self}. \nGateway result: {gateway_result}"
                else:
                    result = f"Failed to send SMS via {self}. \nGateway result: {gateway_result}"
        except Exception as e:
            result = f"Failed to send SMS via {self}.\nException: {e}"

        if not success:
            log_level = logging.ERROR

        return success, log_level, result

    @abstractmethod
    def _send_sms(self, number: str, message: str) -> tuple[bool, str]:
        ...



from pydantic import Field
from enum import Enum

from sms_broker.gateways.config import BaseGatewayConfig


class TeltonikaGatewayConfig(BaseGatewayConfig):
    class Type(str, Enum):
        TELTONIKA = "teltonika"

    type: Type = Field(default=..., title="Type", description="Type of the gateway.")
    ip: str = Field(default=..., title="IP address", description="IP address of the gateway.")
    port: int = Field(default=80, title="Port", description="Port of the gateway.")
    username: str = Field(default="", title="Username", description="Username for authentication.")
    password: str = Field(default="", title="Password", description="Password for authentication.")



import logging
from typing import TYPE_CHECKING

import requests
from pythonping import ping

from sms_broker.gateways.gateway import BaseGateway

if TYPE_CHECKING:
    from sms_broker.gateways.teltonika.config import TeltonikaGatewayConfig

logger = logging.getLogger(__name__)


class TeltonikaGateway(BaseGateway):
    @property
    def config(self) -> "TeltonikaGatewayConfig":
        return super().config

    def _check(self) -> bool:
        result = ping(self._config.ip, self._config.check_timeout, count=self._config.check_retries)

        log_msg = f"packets_loss={result.packet_loss}\n" \
                  f"rtt_avg={result.rtt_avg_ms}\n" \
                  f"rtt_max={result.rtt_max_ms}\n" \
                  f"rtt_min={result.rtt_min_ms}\n" \
                  f"packets_sent={result.stats_packets_sent}\n" \
                  f"packets_received={result.stats_packets_returned}"

        logger.debug(f"Check result for {self}:\n{log_msg}")

        if result.success():
            logger.debug(f"Ping for {self} was successful.")
            return True
        else:
            logger.warning(f"Ping for {self} was unsuccessful.")
            return False

    def _send_sms(self, number: str, message: str) -> tuple[bool, str]:
        response = requests.get(f"http://{self._config.ip}:{self._config.port}/cgi-bin/sms_send",
                                params={"username": self._config.username,
                                        "password": self._config.password,
                                        "number": number,
                                        "text": message},
                                timeout=self._config.timeout)
        gateway_result = response.text.replace("\n", "")
        if response.ok:
            return True, gateway_result
        else:
            return False, gateway_result



from enum import Enum
from ipaddress import IPv4Address, IPv4Network

from pydantic import Field
from sms_broker.server.config import BaseServerConfig


class ApiServerConfig(BaseServerConfig):
    class Type(str, Enum):
        API = "api"

    type: Type = Field(default=..., title="Type", description="Type of the server.")
    host: IPv4Address = Field(default=..., title="API Server Host", description="API Server Host to bind to.")
    port: int = Field(default=..., title="API Server Port", ge=0, le=65535, description="API Server Port to bind to.")
    docs_web_path: str | None = Field(default=None, title="API Server Docs Web Path", description="API Server Docs Web Path.")
    redoc_web_path: str | None = Field(default=None, title="API Server Redoc Web Path", description="API Server Redoc Web Path.")
    allowed_networks: list[IPv4Network] = Field(default_factory=lambda: [IPv4Network("0.0.0.0/0")], title="API Server Allowed Clients Networks",
                                                description="List of allowed client networks.")
    authentication_enabled: bool = Field(default=False, title="API Server Authentication Enabled", description="Enable API Server Authentication.")
    authentication_accounts: dict[str, str] = Field(default_factory=dict, title="API Server Authentication Accounts", description="API Server Authentication Accounts.")
    success_result: str | None = Field(default=None, title="API Server success message",
                                       description="Message to send on success. If set to None, the original message will be sent back to the client.")
    error_result: str | None = Field(default=None, title="API Server error message",
                                     description="Message to send on error. If set to None, the original message will be sent back to the client.")



import time
from typing import Literal

from sms_broker.settings import settings
from wiederverwendbar.typer import Typer

cli_app = Typer(settings=settings)


def main_loop(mode: Literal["listener", "worker"]):
    cli_app.console.print(f"{settings.branding_title} - {mode.capitalize()} is ready. Press CTRL+C to quit.")
    try:
        while True:
            time.sleep(0.001)
    except KeyboardInterrupt:
        cli_app.console.print(f"KeyboardInterrupt received. Stopping {settings.branding_title} - {mode.capitalize()} ...")


@cli_app.command(name="listener", help=f"Start the {settings.branding_title} - listener.")
def listener_command():
    """
    Start the listener.

    :return: None
    """

    from sms_broker.db import db
    from sms_broker.listener import SmsListener

    # print header
    cli_app.console.print(f"Starting {settings.branding_title} - Listener ...")
    cli_app.console.print(f"[white]{cli_app.title_header}[/white]")

    # init db
    db().create_all()

    # start listener
    SmsListener()

    # entering main loop
    main_loop(mode="listener")


@cli_app.command(name="worker", help=f"Start the {settings.branding_title} - worker.")
def worker_command():
    """
    Start the worker.

    :return: None
    """

    from sms_broker.db import db
    from sms_broker.worker import SmsWorker

    # print header
    cli_app.console.print(f"Starting {settings.branding_title} - Worker ...")
    cli_app.console.print(f"[white]{cli_app.title_header}[/white]")

    # init db
    db().create_all()

    # start worker
    SmsWorker()

    # entering main loop
    main_loop(mode="listener")


@cli_app.command(name="init-db", help="Initialize database.")
def init_db_command():
    """
    Initialize database.
    :return: None
    """

    from sms_broker.db import db

    # print header
    cli_app.console.print(f"[white]{cli_app.title_header}[/white]")

    # init db
    cli_app.console.print(f"Initializing database for {settings.branding_title} ...")
    db().create_all()
    cli_app.console.print("Done")



from enum import Enum
from pathlib import Path

from pydantic import Field

from sms_broker.server.config import BaseServerConfig


class FileServerConfig(BaseServerConfig):
    class Type(str, Enum):
        FILE = "file"

    type: Type = Field(default=..., title="Type", description="Type of the server.")
    directory: Path = Field(default=..., title="File Server Directory", description="Directory for incoming files.")
    directory_creating: bool = Field(default=True, title="File Server Creating Directory", description="Create directory if it doesn't exist.")
    directory_scan_interval: int = Field(default=1, title="File Server Scan Interval", description="Scan interval for incoming files.")
    file_extensions: list[str] = Field(default_factory=lambda: [".json"], title="File extensions", description="List of file extensions accepted by the server.")
    file_encoding: str = Field(default="auto", title="TCP Server input encoding", description="Encoding of incoming files.")
    file_delete_on_success: bool = Field(default=True, title="Delete file on success", description="Delete file on success.")
    file_delete_on_error: bool = Field(default=True, title="Delete file on error", description="Delete file on error.")



import logging
from abc import abstractmethod
from datetime import datetime
from threading import Thread
from typing import Any, TYPE_CHECKING, Union

from sms_broker.base import Base
from sms_broker.db import Sms, SmsStatus
from sms_broker.settings import settings

if TYPE_CHECKING:
    from sms_broker.server.config import BaseServerConfig

logger = logging.getLogger(__name__)


class BaseServer(Base, Thread):
    __str_columns__ = ["name",
                       ("debug", "config_host")]

    def __init__(self, name: str, config: "BaseServerConfig"):
        self._is_started = False
        Base.__init__(self, name=name, config=config)
        Thread.__init__(self, name=name, daemon=True)

    @property
    def is_started(self) -> bool:
        return self._is_started

    @property
    def config(self) -> Union["BaseServerConfig", Any]:
        return super().config

    @property
    def config_debug(self) -> bool:
        return self.config.debug

    def run(self):
        logger.info(f"Starting {self} ...")
        try:
            self.enter()
        except KeyboardInterrupt:
            logger.info(f"Stopping {self} ...")
            self.exit()
            logger.debug(f"Stopping {self} ... done")

    @abstractmethod
    def enter(self):
        ...

    def stated_done(self):
        logger.debug(f"Starting {self} ... done.")
        self._is_started = True

    @abstractmethod
    def exit(self):
        ...

    def handle_request(self, caller: Any, **kwargs) -> Any | None:
        logger.debug(f"{self} - Progressing SMS data ...")
        try:
            number, message = self.handle_sms_data(caller=caller, **kwargs)
        except Exception as e:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=e, sms_id=None, result=f"Error while processing SMS body.")
        logger.debug(f"{self} - Progressing SMS data ... done")

        logger.debug(f"Validating SMS ...")

        # check number
        if len(number) > settings.listener.sms_number_max_size:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=False, sms_id=None, result=f"Received number is too long. "
                                                                                                                 f"Max size is '{settings.listener.sms_number_max_size}'.\n"
                                                                                                                 f"number_size={len(number)}")
        _number = ""
        for char in number:
            if char not in list(settings.listener.sms_number_allowed_chars):
                return self.handle_response(caller=self, log_level=logging.ERROR, success=False, sms_id=None, result=f"Received number contains invalid characters. "
                                                                                                                     f"Allowed characters are '{settings.listener.sms_number_allowed_chars}'.\n"
                                                                                                                     f"number='{number}'")
            if char in list(settings.listener.sms_number_replace_chars):
                char = ""
            _number += char
        number = _number
        del _number

        # replace zero number
        if settings.listener.sms_replace_zero_numbers is not None:
            if number.startswith("0"):
                number = settings.listener.sms_replace_zero_numbers + number[1:]

        # check a message
        if len(message) > settings.listener.sms_message_max_size:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=False, sms_id=None, result=f"Received message is too long. "
                                                                                                                 f"Max size is '{settings.listener.sms_message_max_size}'.\n"
                                                                                                                 f"message_size={len(message)}")

        logger.debug(f"Validating SMS ... done")

        # log sms
        if settings.listener.sms_logging:
            logger.debug(f"SMS:\nnumber={number}\nmessage='{message}'")

        # queue sms
        logger.info(f"Queuing SMS ...")
        try:
            sms = Sms(status=SmsStatus.QUEUED,
                      received_by=self.name,
                      received_datetime=datetime.now(),
                      number=number,
                      message=message)
            sms.save()
            success = True
            sms_id = sms.id
            result = f"SMS with id={sms_id} queued successfully."
        except Exception as e:
            success = False
            sms_id = None
            result = f"Error while queuing SMS."
            if self.config.debug:
                result += f"\nException: {e}"
        logger.log(logging.DEBUG if success else logging.ERROR, f"Queuing SMS ... {'done' if success else 'failed'} --> {result}")

        # send response
        if success:
            log_level = logging.DEBUG
        else:
            log_level = logging.ERROR
        return self.handle_response(caller=caller, log_level=log_level, success=success, sms_id=sms_id, result=result, **kwargs)

    def handle_response(self, caller: Any, log_level: int, success: bool | Exception, sms_id: int | None, result: str, **kwargs) -> Any | None:
        if result.endswith(".") or result.endswith(":"):
            result = result[:-1]
        e = ""
        if isinstance(success, Exception):
            e = str(success)
            success = False
        if not success and self.config.debug:
            result = f"{result}:\n{e}"
        else:
            result = f"{result}."
        if not success:
            log_msg = f"{self} - {result}:\n{e}"
        else:
            log_msg = f"{self} - {result}."
        logger.log(log_level, log_msg)

        logger.debug(f"{self} - Sending Response.\nsuccess='{success}'\nsms_id={sms_id}\nresult={result}")
        try:
            self.increase_sms_count()
            if not success:
                self.increase_sms_error_count()
                return self.error_handler(caller=caller, sms_id=sms_id, result=result, **kwargs)
            return self.success_handler(caller=caller, sms_id=sms_id, result=result, **kwargs)
        except Exception as e:
            logger.error(f"{self} - Error while sending response message.\n{e}")
        return None

    @abstractmethod
    def handle_sms_data(self, caller: Any, **kwargs) -> tuple[str, str]:
        ...

    @abstractmethod
    def success_handler(self, caller: Any, sms_id: int, result: str, **kwargs) -> Any:
        ...

    @abstractmethod
    def error_handler(self, caller: Any, sms_id: int | None, result: str, **kwargs) -> Any:
        ...



from enum import Enum
from ipaddress import IPv4Address, IPv4Network

from pydantic import Field

from sms_broker.server.config import BaseServerConfig


class TcpServerConfig(BaseServerConfig):
    class Type(str, Enum):
        TCP = "tcp"

    type: Type = Field(default=..., title="Type", description="Type of the server.")
    host: IPv4Address = Field(default=..., title="TCP Server Host", description="TCP Server Host to bind to.")
    port: int = Field(default=..., title="TCP Server Port", ge=0, le=65535, description="TCP Server Port to bind to.")
    allowed_networks: list[IPv4Network] = Field(default_factory=lambda: [IPv4Network("0.0.0.0/0")], title="TCP Server Allowed Clients Networks",
                                                description="List of allowed client networks.")
    in_encoding: str = Field(default="auto", title="TCP Server input encoding", description="Encoding of incoming data.")
    out_encoding: str = Field(default="utf-8", title="TCP Server output encoding", description="Encoding of outgoing data.")
    data_max_size: int = Field(default=2048, title="Max Data Size", description="Max Data Size for SMS.", ge=1024)
    success_result: str | None = Field(default=None, title="TCP Server success message",
                                       description="Message to send on success. If set to None, the original message will be sent back to the client.")
    error_result: str | None = Field(default=None, title="TCP Server error message",
                                     description="Message to send on error. If set to None, the original message will be sent back to the client.")



import logging
import socketserver
import sys
from ipaddress import IPv4Address
from typing import TYPE_CHECKING, Any

import chardet

from sms_broker.server.server import BaseServer

if TYPE_CHECKING:
    from sms_broker.server.tcp.config import TcpServerConfig

logger = logging.getLogger(__name__)


class TcpServerHandler(socketserver.BaseRequestHandler):
    server: "TcpServer"

    def handle(self) -> None:
        # get client ip and port
        client_ip, client_port = self.client_address
        try:
            client_ip = IPv4Address(client_ip)
        except Exception as e:
            return self.server.handle_response(caller=self, log_level=logging.ERROR, success=e, sms_id=None, result=f"Error while parsing client IP address.")

        self.server.handle_request(caller=self, client_ip=client_ip, client_port=client_port)
        return None


class TcpServer(BaseServer, socketserver.TCPServer):
    __str_columns__ = ["name",
                       ("debug", "config_debug"),
                       ("host", "config_host"),
                       ("port", "config_port"),
                       ("allowed_networks", "config_allowed_networks")]

    def __init__(self, name: str, config: "TcpServerConfig"):
        BaseServer.__init__(self, name=name, config=config)

        try:
            # noinspection PyTypeChecker
            socketserver.TCPServer.__init__(self, (str(self.config.host), self.config.port), TcpServerHandler)
        except Exception as e:
            logger.error(f"Error while initializing {self}: {e}")
            sys.exit(1)

        self.init_done()

    @property
    def config(self) -> "TcpServerConfig":
        return super().config

    @property
    def config_host(self) -> str:
        return str(self.config.host)

    @property
    def config_port(self) -> int:
        return self.config.port

    @property
    def config_allowed_networks(self) -> list[str]:
        return [str(allowed_network) for allowed_network in self.config.allowed_networks]

    def enter(self):
        self.stated_done()
        self.serve_forever()

    def exit(self):
        self.shutdown()
        self.server_close()

    # noinspection DuplicatedCode
    def handle_request(self, caller: Any, **kwargs) -> Any | None:
        # check if client ip is allowed
        allowed = False
        for network in self.config.allowed_networks:
            if kwargs["client_ip"] in network:
                allowed = True
                break
        if not allowed:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=False, sms_id=None, result=f"Client IP address '{kwargs['client_ip']}' is not allowed.")

        logger.debug(f"{self} - Accept message:\nclient='{kwargs['client_ip']}'\nport={kwargs['client_ip']}")

        return super().handle_request(caller=caller, **kwargs)

    def handle_sms_data(self, caller: TcpServerHandler, **kwargs) -> tuple[str, str] | None:
        # get data
        try:
            data = caller.request.recv(self.config.data_max_size).strip()
            logger.debug(f"{self} - data={data}")
        except Exception as e:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=e, sms_id=None, result=f"Error while receiving data.")

        # detect encoding
        try:
            encoding = self.config.in_encoding
            if encoding == "auto":
                encoding = chardet.detect(data)['encoding']
            logger.debug(f"{self} - encoding={encoding}")
        except Exception as e:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=e, sms_id=None, result=f"Error while detecting encoding.")

        # decode message
        try:
            data_str = data.decode(encoding)
            logger.debug(f"{self} - data_str='{data_str}'")

            # split message
            if "\r\n" not in data_str:
                return self.handle_response(caller=self, log_level=logging.ERROR, success=False, sms_id=None, result=f"Received data is not valid.")
            number, message = data_str.split("\r\n")
            return number, message
        except Exception as e:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=e, sms_id=None, result=f"Error while decoding data.")

    def success_handler(self, caller: TcpServerHandler, sms_id: int, result: str, **kwargs) -> Any:
        if self.config.success_result is not None:
            result = self.config.success_result
        result_raw = result.encode(self.config.out_encoding)
        caller.request.sendall(result_raw)

    def error_handler(self, caller: TcpServerHandler, sms_id: int | None, result: str, **kwargs) -> Any:
        if self.config.error_result is not None:
            result = self.config.error_result
        result_raw = result.encode(self.config.out_encoding)
        caller.request.sendall(result_raw)



from pydantic import Field

from sms_broker.base import BaseConfig


class BaseServerConfig(BaseConfig):
    debug: bool = Field(default=False, title="Debug", description="If set to True, server will be started in debug mode.")



from enum import Enum

from pydantic import Field

from sms_broker.gateways.config import BaseGatewayConfig


class VonageGatewayConfig(BaseGatewayConfig):
    class Type(str, Enum):
        VONAGE = "vonage"

    type: Type = Field(default=..., title="Type", description="Type of the gateway.")
    api_key: str = Field(default="", title="API key", description="API key for authentication.")
    api_secret: str = Field(default="", title="API secret", description="API secret for authentication.")
    from_text: str = Field(default="SMS-Broker", title="From Text", description="From Text visible for recipient.")
    check_balance: bool = Field(default=True, title="Check balance", description="If set to True, balance will be checked before sending SMS.")
    check_min_balance: float = Field(default=0.0, title="Check min balance", description="Minimum balance required for checking gateway availability.")
    check_auto_balance: bool = Field(default=True, title="Check auto balance",
                                     description="If set to True, min balance will be ignored if the vonage api returns an auto reload flag.")



import logging
from contextlib import asynccontextmanager
from ipaddress import IPv4Address
from pathlib import Path
from typing import TYPE_CHECKING, Any

import uvicorn
from fastapi import FastAPI, Request, Response
from fastapi.responses import RedirectResponse
from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.routing import Route
from starlette_admin import fields as sa_fields
from starlette_admin.auth import AdminConfig, AdminUser, AuthProvider
from starlette_admin.exceptions import LoginFailed, FormValidationError
from starlette_admin.actions import row_action, action
from starlette_admin.contrib.sqla import Admin, ModelView
from starlette_admin.exceptions import ActionFailed

from sms_broker.statics import ASSETS_PATH
from sms_broker.db import Sms, SmsStatus, db
from sms_broker.server.server import BaseServer
from sms_broker.settings import settings

if TYPE_CHECKING:
    from sms_broker.server.ui.config import UiServerConfig

logger = logging.getLogger(__name__)


class UiAuthProvider(AuthProvider):
    ui: "Ui"

    async def login(self,
                    username: str,
                    password: str,
                    remember_me: bool,
                    request: Request,
                    response: Response) -> Response:
        if username not in self.ui.ui_server.config.authentication_accounts:
            if self.ui.ui_server.config.debug:
                raise FormValidationError({"username": "Username not found!"})
            raise LoginFailed("Invalid username or password!")
        if self.ui.ui_server.config.authentication_accounts[username] != password:
            if self.ui.ui_server.config.debug:
                raise FormValidationError({"password": "Password incorrect!"})
            raise LoginFailed("Invalid username or password!")

        # save username in session
        request.session.update({"username": username})
        return response

    async def is_authenticated(self, request) -> bool:
        if request.session.get("username", None) in self.ui.ui_server.config.authentication_accounts:
            return True
        return False

    def get_admin_config(self, request: Request) -> AdminConfig:
        return AdminConfig(
            app_title=f"Hello, {request.session['username'].capitalize()}!",
            logo_url=None,
        )

    def get_admin_user(self, request: Request) -> AdminUser:
        return AdminUser(username=request.session["username"], photo_url=None)

    async def logout(self, request: Request, response: Response) -> Response:
        request.session.clear()
        return response


class UiSmsView(ModelView):
    page_size = 100
    page_size_options = [5, 10, 25, 50, 75, 100, -1]
    fields = [sa_fields.IntegerField("id"),
              sa_fields.EnumField("status", choices=[status.value for status in SmsStatus]),
              sa_fields.StringField("received_by"),
              sa_fields.DateTimeField("received_datetime"),
              sa_fields.DateTimeField("processed_datetime"),
              sa_fields.StringField("sent_by"),
              sa_fields.PhoneField("number"),
              sa_fields.TextAreaField("message"),
              sa_fields.TextAreaField("result"),
              sa_fields.TextAreaField("log")]
    exclude_fields_from_list = [Sms.message,
                                Sms.result,
                                Sms.log]
    exclude_fields_from_create = [Sms.status,
                                  Sms.received_by,
                                  Sms.received_datetime,
                                  Sms.processed_datetime,
                                  Sms.sent_by,
                                  Sms.result,
                                  Sms.log]

    row_actions = ["view", "row_reset", "row_abort"]
    actions = ["reset", "abort"]

    def __init__(self, ui: "Ui"):
        if not settings.listener.sms_logging:
            # noinspection PyUnresolvedReferences
            self.exclude_fields_from_detail.append(Sms.message)
        super().__init__(Sms, label="SMS", icon="fa fa-message")
        self.ui = ui

    def can_edit(self, request: Request) -> bool:
        return False

    def can_delete(self, request: Request) -> bool:
        return False

    async def create(self, request: Request, data: dict[str, Any]) -> Any:
        try:
            data = await self._arrange_data(request, data)
            await self.validate(request, data)

            # get number and message
            number = data["number"]
            message = data["message"]

            client_ip = IPv4Address(request.client.host)
            client_port = request.client.port

            result = self.ui.ui_server.handle_request(caller=None, number=number, message=message, client_ip=client_ip, client_port=client_port)
            if isinstance(result, Exception):
                raise result
            sms = Sms.get(id=result)
            if sms is None:
                raise FileNotFoundError(f"SMS with id={result} not found!")
            return sms
        except Exception as e:
            return self.handle_exception(e)

    @row_action(
        name="row_reset",
        text="Reset SMS",
        confirmation="Do you want to reset this SMS?",
        icon_class="fa-regular fa-repeat",
        submit_btn_text="Yes",
        submit_btn_class="btn-success",
    )
    async def row_reset_action(self, request: Request, pk: str) -> str:
        sms = Sms.get(Sms.id == pk)
        if sms is None:
            raise ActionFailed(f"SMS with id={pk} not found.")
        sms.update(status=SmsStatus.QUEUED,
                   processed_datetime=None,
                   sent_by=None,
                   result=None,
                   log=None)
        return f"SMS with id={pk} reset successfully."

    @action(
        name="reset",
        text="Reset SMS",
        confirmation="Do you want to reset this SMS?",
        icon_class="fa-regular fa-repeat",
        submit_btn_text="Yes",
        submit_btn_class="btn-success",
    )
    async def reset_action(self, request: Request, pks: list[str]) -> str:
        successes = []
        for pk in pks:
            successes.append(await self.row_reset_action(request=request, pk=pk))
        return "\n".join(successes)

    @row_action(
        name="row_abort",
        text="Abort SMS",
        confirmation="Do you want to cancel this SMS?",
        icon_class="fa-regular fa-ban",
        submit_btn_text="Yes",
        submit_btn_class="btn-success",
    )
    async def row_abort_action(self, request: Request, pk: str) -> str:
        sms = Sms.get(Sms.id == pk)
        if sms is None:
            raise ActionFailed(f"SMS with id={pk} not found.")
        if sms.status != SmsStatus.QUEUED:
            raise ActionFailed(f"Cannot abort SMS with id={pk}! SMS is not in queued state.")
        sms.update(status=SmsStatus.ABORTED,
                   processed_datetime=None,
                   sent_by=None,
                   result=None,
                   log=None)
        return f"SMS with id={pk} aborted successfully."

    @action(
        name="abort",
        text="Abort SMS",
        confirmation="Do you want to cancel this SMS?",
        icon_class="fa-regular fa-ban",
        submit_btn_text="Yes",
        submit_btn_class="btn-success",
    )
    async def abort_action(self, request: Request, pks: list[str]) -> str:
        successes = []
        for pk in pks:
            successes.append(await self.row_abort_action(request=request, pk=pk))
        return "\n".join(successes)


class Ui(Admin):
    def __init__(self, ui_server: "UiServer"):
        templates_dir = Path(__file__).parent / "templates"
        if not templates_dir.is_dir():
            raise FileNotFoundError(f"Template directory '{templates_dir}' not found.")
        auth_provider = UiAuthProvider(allow_paths=["/statics/logo.png"])
        auth_provider.ui = self
        super().__init__(engine=db().engine,
                         title=ui_server.title,
                         base_url="/",
                         templates_dir=str(templates_dir),
                         statics_dir=ASSETS_PATH,
                         logo_url="/statics/logo.png",
                         login_logo_url="/statics/logo.png",
                         favicon_url="/statics/favicon.ico",
                         auth_provider=auth_provider,
                         middlewares=[Middleware(SessionMiddleware,
                                                 secret_key=ui_server.config.session_secret_key,
                                                 session_cookie=ui_server.config.session_cookie,
                                                 max_age=ui_server.config.session_max_age)],
                         debug=ui_server.config.debug)
        self.ui_server = ui_server

        # add views
        self.sms_view = UiSmsView(self)
        self.add_view(self.sms_view)

        # replace the root route
        root_route_index = None
        for i, route in enumerate(self.routes):
            if not isinstance(route, Route):
                continue
            if route.path != "/":
                continue
            root_route_index = i
            break
        if root_route_index is None:
            raise RuntimeError("Root route not found.")
        self.routes[root_route_index] = Route(
            path="/",
            endpoint=self.root,
            methods=None,
            name="index",
        )

        # mount to ui_server
        self.mount_to(self.ui_server)

    async def root(self, request: Request):
        return RedirectResponse(url=f"/{self.sms_view.identity}/list")


class UiServer(BaseServer, FastAPI):
    __str_columns__ = ["name",
                       ("debug", "config_debug"),
                       ("host", "config_host"),
                       ("port", "config_port"),
                       ("allowed_networks", "config_allowed_networks"),
                       ("authentication_enabled", "config_authentication_enabled")]

    def __init__(self, name: str, config: "UiServerConfig"):
        BaseServer.__init__(self,
                            name=name,
                            config=config)
        FastAPI.__init__(self,
                         openapi_url=None,
                         docs_url=None,
                         redoc_url=None,
                         lifespan=self._stated_done,
                         debug=self.config.debug)
        self.title = f"{settings.branding_title} - {self.name}"
        self.description = settings.branding_description
        self.version = f"v{settings.branding_version}"
        self.terms_of_service = settings.branding_terms_of_service
        self.contact = {"name": settings.branding_author, "email": settings.branding_author_email}
        self.license_info = {"name": settings.branding_license, "url": settings.branding_license_url}

        # create ui
        logger.info(f"Create ui for {self} ...")
        self._ui = Ui(self)
        logger.debug(f"Create ui for {self} ... done")

        self.init_done()

    @property
    def config(self) -> "UiServerConfig":
        return super().config

    @property
    def config_host(self) -> str:
        return str(self.config.host)

    @property
    def config_port(self) -> int:
        return self.config.port

    @property
    def config_allowed_networks(self) -> list[str]:
        return [str(allowed_network) for allowed_network in self.config.allowed_networks]

    @property
    def config_authentication_enabled(self) -> bool:
        return self.config.authentication_enabled

    @staticmethod
    @asynccontextmanager
    async def _stated_done(ui_server: "UiServer"):
        ui_server.stated_done()
        yield

    def enter(self):
        uvicorn.run(self, host=str(self.config.host), port=self.config.port)

    def exit(self):
        ...

    # noinspection DuplicatedCode
    def handle_request(self, caller: None, **kwargs) -> Any | None:
        # check if client ip is allowed
        allowed = False
        for network in self.config.allowed_networks:
            if kwargs["client_ip"] in network:
                allowed = True
                break
        if not allowed:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=False, sms_id=None, result=f"Client IP address '{kwargs['client_ip']}' is not allowed.")

        logger.debug(f"{self} - Accept message:\nclient='{kwargs['client_ip']}'\nport={kwargs['client_ip']}")

        return super().handle_request(caller=caller, **kwargs)

    def handle_sms_data(self, caller: None, **kwargs) -> tuple[str, str]:
        return kwargs["number"], kwargs["message"]

    def success_handler(self, caller: None, sms_id: int, result: str, **kwargs) -> Any:
        return sms_id

    def error_handler(self, caller: None, sms_id: int | None, result: str, **kwargs) -> Any:
        return RuntimeError(result)



from pydantic import Field

from sms_broker.base import BaseConfig


class BaseGatewayConfig(BaseConfig):
    dry_run: bool = Field(default=False, title="Dry run mode", description="If set to True, SMS will not be sent via this gateway."
                                                                           "This is useful for testing purposes.")
    timeout: int = Field(default=5, title="Timeout", description="Timeout for sending SMS via this gateway.")
    check: bool = Field(default=True, title="Check", description="If set to True, gateway will be checked before sending SMS.")
    check_timeout: int = Field(default=1, title="Check timeout", description="Timeout for checking gateway availability.")
    check_retries: int = Field(default=3, title="Check retries", description="Number of retries for checking gateway availability.")



from enum import Enum
from datetime import datetime

from sqlalchemy import Column, Integer, DateTime, VARCHAR

from wiederverwendbar.singleton import Singleton
from wiederverwendbar.sqlalchemy import SqlalchemyDb, Base, EnumValueStr

from sms_broker.settings import settings


class Db(SqlalchemyDb, metaclass=Singleton):
    ...


def db() -> Db:
    try:
        return Singleton.get_by_type(Db)
    except RuntimeError:
        # noinspection PyArgumentList
        return Db(settings=settings, init=True)


class SmsStatus(str, Enum):
    QUEUED = "queued"
    SENT = "sent"
    ABORTED = "aborted"
    ERROR = "error"


class Sms(Base, db().Base):
    __tablename__ = "sms"
    __str_columns__ = ["id", "status", "received_datetime", "send_datetime", "number"]

    id: int = Column(Integer(), primary_key=True, autoincrement=True, name="sms_id")
    status: SmsStatus = Column(EnumValueStr(SmsStatus), nullable=False, name="sms_status")
    received_by: str = Column(VARCHAR(20), nullable=False, name="sms_received_by")
    received_datetime: datetime = Column(DateTime(), nullable=False, name="sms_received_datetime")
    processed_datetime: datetime = Column(DateTime(), nullable=True, name="sms_processed_datetime")
    sent_by: str = Column(VARCHAR(20), nullable=True, name="sms_sent_by")
    number: str = Column(VARCHAR(50), nullable=False, name="sms_number")
    message: str = Column(VARCHAR(1600), nullable=False, name="sms_message")
    result: str = Column(VARCHAR(1000), nullable=True, name="sms_result")
    log: str = Column(VARCHAR(10000), nullable=True, name="sms_log")



import json
import logging
from typing import Literal, TYPE_CHECKING

from vonage import Vonage, Auth, HttpClientOptions
from vonage_sms import SmsMessage, SmsResponse

from sms_broker.gateways.gateway import BaseGateway

if TYPE_CHECKING:
    from sms_broker.gateways.vonage.config import VonageGatewayConfig

logger = logging.getLogger(__name__)


class VonageGateway(BaseGateway):
    @property
    def config(self) -> "VonageGatewayConfig":
        return super().config

    def get_vonage_instance(self, mode: Literal["check", "send"]) -> Vonage:
        # Create an Auth instance
        auth = Auth(api_key=self._config.api_key, api_secret=self._config.api_secret)

        # Create HttpClientOptions instance
        if mode == "check":
            options = HttpClientOptions(timeout=self._config.check_timeout, max_retries=self._config.check_retries)
        elif mode == "send":
            options = HttpClientOptions(timeout=self._config.timeout)
        else:
            raise ValueError("Invalid mode")

        # Create a Vonage instance
        vonage = Vonage(auth=auth, http_client_options=options)

        return vonage

    def _check(self) -> bool:
        vonage = self.get_vonage_instance(mode="check")
        if not self.config.check_balance:
            return True
        balance = vonage.account.get_balance()
        log_msg = f"balance={balance.value}\n" \
                  f"auto_reload={balance.auto_reload}"

        logger.debug(f"Checking account balance for {self}:\n{log_msg}")

        if balance.value >= self._config.check_min_balance:
            logger.debug(f"Balance for {self} is enough.")
            return True
        if self._config.check_auto_balance:
            if balance.auto_reload:
                logger.debug(f"Balance for {self} is not enough. But auto_reload is enabled.")
                return True
            logger.warning(f"Balance for {self} is not enough and auto_reload is disabled.")
        else:
            logger.warning(f"Balance for {self} is not enough.")
        return False

    def _send_sms(self, number: str, message: str) -> tuple[bool, str]:
        vonage = self.get_vonage_instance(mode="send")
        message = SmsMessage(to=number, from_=self._config.from_text, text=message, **{})
        sms_response: SmsResponse = vonage.sms.send(message)

        # check status for all messages
        success = True
        for message in sms_response.messages:
            if message.status != "0":
                success = False
                break

        # convert sms_response to pretty json
        sms_response_dict = []
        for message in sms_response.messages:
            sms_response_dict.append(message.model_dump(mode="json"))
        sms_response_json = json.dumps(sms_response_dict, indent=4)

        return success, sms_response_json



__title__ = "SMS Broker"
__description__ = "A broker server for sending SMS."
__version__ = "3.3.0"
__author__ = "Julius Koenig"
__author_email__ = "info@bastelquartier.de"
__license__ = "GPL-3.0"
__license_url__ = "https://www.gnu.org/licenses/gpl-3.0.md"
__terms_of_service__ = "https://bastelquartier.de/terms-of-service"



import os
from pathlib import Path
from typing import Union

from pydantic import Field
from wiederverwendbar.sqlalchemy import SqlalchemySettings

from wiederverwendbar.logger import LoggerSettings
from wiederverwendbar.pydantic import FileConfig
from wiederverwendbar.typer import TyperSettings

from sms_broker import __title__, __description__, __author__, __author_email__, __version__, __license__, __license_url__, __terms_of_service__
from sms_broker.server.file.config import FileServerConfig
from sms_broker.server.tcp.config import TcpServerConfig
from sms_broker.server.api.config import ApiServerConfig
from sms_broker.server.ui.config import UiServerConfig
from sms_broker.gateways.teltonika.config import TeltonikaGatewayConfig
from sms_broker.gateways.vonage.config import VonageGatewayConfig

ENV_PREFIX = "SMS_BROKER_"

AVAILABLE_SERVER_CONFIGS = Union[
    FileServerConfig,
    TcpServerConfig,
    ApiServerConfig,
    UiServerConfig
]

AVAILABLE_GATEWAY_CONFIGS = Union[
    TeltonikaGatewayConfig,
    VonageGatewayConfig
]


class Settings(FileConfig, TyperSettings, SqlalchemySettings):
    # listener
    class ListenerSettings(LoggerSettings):
        # sms
        sms_number_allowed_chars: str = Field(default="+*#()0123456789 ", title="Allowed Number Characters", description="Allowed Number Characters.")
        sms_number_replace_chars: str = Field(default="() ", title="Replace Number Characters", description="Replace Number Characters.")
        sms_replace_zero_numbers: str | None = Field(default=None, title="Replace Zero Numbers",
                                                     description="Replace all zero numbers with this string.")
        sms_number_max_size: int = Field(default=20, title="Max Number Size", description="Max Number Size for SMS.", ge=1, le=50)
        sms_message_max_size: int = Field(default=1600, title="Max Message Size", description="Max Message Size for SMS.", ge=1, le=1600)
        sms_logging: bool = Field(default=False, title="SMS Logging", description="Enable SMS Logging content logging.")

        # server
        server: dict[str, AVAILABLE_SERVER_CONFIGS] = Field(default_factory=dict, title="Server",
                                                            description="Server configuration.")

    listener: ListenerSettings = Field(default_factory=ListenerSettings, title="Listener Settings", description="Listener settings.")

    # worker
    class WorkerSettings(LoggerSettings):
        # worker
        count: int | None = Field(default=None, title="Background Worker Count",
                                  description="Number of background workers. If None, worker count will be calculated by cpu count but max 4.")

        # sms
        sms_handle_interval: int = Field(default=1, title="SMS handle Interval", description="Interval for handling SMS in seconds.")
        sms_cleanup_max_age: int | None = Field(default=60 * 60 * 24 * 30, title="DB SMS cleanup max age",
                                                description="Time after cleanup SMS from DB in seconds. If None, no cleanup will be performed.")
        sms_cleanup_interval: int = Field(default=60, title="DB SMS cleanup interval", description="Interval for cleanup SMS from DB in seconds.")

        # gateways
        gateways: dict[str, AVAILABLE_GATEWAY_CONFIGS] = Field(default_factory=dict, title="Gateways",
                                                               description="Gateways configuration.")

    worker: WorkerSettings = Field(default_factory=WorkerSettings, title="Worker Settings", description="Worker settings.")


# noinspection PyArgumentList
settings = Settings(file_path=Path(os.environ.get(f"{ENV_PREFIX}SETTINGS_FILE", "settings.json")),
                    file_must_exist="yes_print",
                    branding_title=__title__,
                    branding_description=__description__,
                    branding_version=__version__,
                    branding_author=__author__,
                    branding_author_email=__author_email__,
                    branding_license=__license__,
                    branding_license_url=__license_url__,
                    branding_terms_of_service=__terms_of_service__,)


import logging
import sys
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any

import chardet
from pydantic import BaseModel, Field

from sms_broker.server.server import BaseServer

if TYPE_CHECKING:
    from sms_broker.server.file.config import FileServerConfig

logger = logging.getLogger(__name__)


class FileModel(BaseModel):
    number: str = Field(default=..., title="Number", description="The phone number of the SMS.")
    message: str = Field(default=..., title="Message", description="The message of the SMS.")


class FileServer(BaseServer):
    __str_columns__ = ["name",
                       ("debug", "config_debug"),
                       ("directory", "config_directory")]

    def __init__(self, name: str, config: "FileServerConfig"):
        BaseServer.__init__(self, name=name, config=config)

        self.init_done()

    @property
    def config(self) -> "FileServerConfig":
        return super().config

    @property
    def config_directory(self) -> str:
        return str(self.config.directory.absolute())

    def enter(self):
        # check if directory exist
        if not self.config.directory.is_dir():
            if not self.config.directory_creating:
                logger.error(f"Directory '{self.config.directory}' does not exist.")
                sys.exit(1)
            try:
                logger.info(f"Created directory '{self.config.directory}' ...")
                self.config.directory.mkdir(parents=True)
                logger.info(f"Created directory '{self.config.directory}' ... done")
            except Exception as e:
                logger.error(f"Error while creating directory '{self.config.directory}': {e}")
                sys.exit(1)

        self.stated_done()

        while True:
            last = time.perf_counter()

            # iterate files from directory
            for file_path in self.config.directory.iterdir():
                if file_path.suffix not in self.config.file_extensions:
                    continue
                return self.handle_request(caller=None, file_path=file_path)

            # wait loop intervall
            while time.perf_counter() - last < self.config.directory_scan_interval:
                time.sleep(0.01)

    def exit(self):
        ...

    # noinspection DuplicatedCode
    def handle_request(self, caller: None, **kwargs) -> Any | None:
        logger.debug(f"{self} - Accept message:\nfile_path='{kwargs['file_path']}'")
        return super().handle_request(caller=caller, **kwargs)

    def handle_sms_data(self, caller: None, **kwargs) -> tuple[str, str] | None:
        # read file
        file_path: Path = kwargs["file_path"]
        try:
            with file_path.open("rb") as file:
                data  = file.read()
            logger.debug(f"{self} - data={data}")
        except Exception as e:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=e, sms_id=None, result=f"Error while reading data.")

        # detect encoding
        try:
            encoding = self.config.file_encoding
            if encoding == "auto":
                encoding = chardet.detect(data)['encoding']
            logger.debug(f"{self} - encoding={encoding}")
        except Exception as e:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=e, sms_id=None, result=f"Error while detecting encoding.")

        # decode message
        try:
            data_str = data.decode(encoding)
            logger.debug(f"{self} - data_str='{data_str}'")
        except Exception as e:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=e, sms_id=None, result=f"Error while decoding data.")

        # parse to FileModel
        try:
            file_model = FileModel.model_validate_json(data_str)
            logger.debug(f"{self} - file_model={file_model}")
        except Exception as e:
            return self.handle_response(caller=self, log_level=logging.ERROR, success=e, sms_id=None, result=f"Error while parsing data.")

        return file_model.number, file_model.message

    def success_handler(self, caller: None, sms_id: int, result: str, **kwargs) -> Any:
        file_path: Path = kwargs["file_path"]
        if self.config.file_delete_on_success:
            file_path.unlink()

    def error_handler(self, caller: None, sms_id: int | None, result: str, **kwargs) -> Any:
        file_path: Path = kwargs["file_path"]
        if self.config.file_delete_on_error:
            file_path.unlink()



import logging
import warnings
from abc import ABC
from typing import Any, Union

from pydantic import BaseModel, PrivateAttr
from wiederverwendbar.functions.get_pretty_str import get_pretty_str

logger = logging.getLogger(__name__)


class Base(ABC):
    __str_columns__: list[str | tuple[str]] = ["name"]

    def __init__(self, name: str, config: "BaseConfig"):
        self._name = name
        self._config = config
        self._sms_count = 0
        self._sms_error_count = 0

        logger.info(f"Initializing {self} ...")

    def __str__(self):
        out = f"{self.__class__.__name__}("
        for attr_name in self.__str_columns__:
            if type(attr_name) is tuple:
                attr_view_name = attr_name[0]
                attr_name = attr_name[1]
            else:
                attr_view_name = attr_name
            if not hasattr(self, attr_name):
                warnings.warn(f"Attribute '{attr_name}' is not set for {self}.")
            out += f"{attr_view_name}={get_pretty_str(getattr(self, attr_name))}, "
        out = out[:-2] + ")"
        return out

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name}, config={self.config})"

    @property
    def name(self) -> str:
        return self._name

    @property
    def config(self) -> Union["BaseConfig", Any]:
        return self._config

    @property
    def sms_count(self) -> int:
        return self._sms_count

    @property
    def sms_error_count(self) -> int:
        return self._sms_error_count

    def init_done(self):
        logger.debug(f"Initializing {self} ... done")

    def increase_sms_count(self) -> None:
        self._sms_count += 1

    def increase_sms_error_count(self) -> None:
        self._sms_error_count += 1

    def reset_metrics(self):
        self._sms_count = 0
        self._sms_error_count = 0

class BaseConfig(BaseModel):
    class Config:
        use_enum_values = True
